#ifndef _MYTMPLLIB_H
#define _MYTMPLLIB_H

#include <string>
#include <new>

using namespace std;

struct MyTmpltLibXptn
{
   enum {  paramError, noArray, allocFailed,
           cannotCreate, cannotWrite  };

   char funcName[100];
   int  errCode;
   int  errIntVal1, errIntVal2;
   char errStrVal[100];
   MyTmpltLibXptn( const char* fn, int ec,
                   int erv1, int erv2 = 0 )
   {  strncpy( funcName, fn, 99 );  funcName[99] = '\0';
      errCode = ec;
      errIntVal1 = erv1;  errIntVal2 = erv2;  }
   MyTmpltLibXptn( const char* fn, int ec, const char* sv="" )
   {  strncpy( funcName, fn, 99 );  funcName[99] = '\0';
      errCode = ec;
      strncpy( errStrVal, sv, 99 );  errStrVal[99] = '\0';  }
}; // MyTmpltLibXptn

// linSearch --    v[from],... v[upto] 
//                 x.   
//                        
//              
//                   -1
template< typename T >
int linSearch( const T v[], int n,
               int from, int upto, const T& x )
{
   if ( v == 0 && n > 0 )
      throw MyTmpltLibXptn( "linSearch",
                            MyTmpltLibXptn::noArray );
   int fv( -1 );
   if ( v != 0 && (0 <= from && from <= upto && upto < n) )
   {
      T* ncv( const_cast<T*>(v) );
      T  save( v[upto+1] );  //   v[upto+1]
      ncv[upto+1] = x;       // 
      int k( from );
      while ( v[k] != x )  ++k;
      if ( k <= upto )  fv = k;
                  else  fv = -1;
      ncv[upto+1] = save;    //    
   // (from <= fv <= upto && v[fv] == x) ||
   //                (fv == -1 && ( j:from..upto o v[j] != x))
   }
   return fv;
} // linSearch

// renew       nf   T 
//           ni (<= nf)       ni
//            p.
template< typename T >
void renew( T*& p, int ni, int nf )
{
   if ( ni < 0 || nf < ni ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::paramError, ni, nf );
// 0 <= ni <= nf      
   if ( p == 0 && ni > 0 ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::noArray );
// (0 <= ni <= nf) && (p != 0 || ni == 0)
   T* temp;
   try
   {  temp = new T[nf];  }
   catch( ... )
   {  throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   try
   {
      for ( int k(0); k < ni; ++k )  temp[k] = p[k]; 
   }
   catch( ... )
   {  delete[] temp;
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   delete[] p;   p = temp;
} // void renew
/*
// renew       nf   T 
//           ni (<= nf)       ni
//            p.
template< typename T >
void renew( T*& p, int ni, int nf )
{
   if ( ni < 0 || nf < ni ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::paramError,
                            ni, nf );
// 0 <= ni <= nf      
   if ( p == 0 && ni > 0 ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::noArray );
// (0 <= ni <= nf) && (p != 0 || ni == 0)
   T* temp;
   try
   {  temp = new T[nf];  }
   catch( bad_alloc& )
   {  throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   try
   {
      for ( int k(0); k < ni; ++k )  temp[k] = p[k]; 
   }
   catch( bad_alloc& )
   {  delete[] temp;
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   delete[] p;   p = temp;
} // void renew

// renew       nf   T 
//           ni (<= nf)       ni
//            p.
//          TXptn      T    
//           allocFailed 
template< typename T, typename TXptn >
void renew( T*& p, int ni, int nf )
{
   if ( ni < 0 || nf < ni ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::paramError,
                            ni, nf );
// 0 <= ni <= nf      
   if ( p == 0 && ni > 0 ) 
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::noArray );
// (0 <= ni <= nf) && (p != 0 || ni == 0)
   T* temp;
   try
   {  temp = new T[nf];  }
   catch( bad_alloc& )
   {  throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   catch( TXptn& x )
   {
      if ( x.errorCode == TXptn::allocFailed )
         throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed);
      else
         throw;
   }
   try
   {
      for ( int k(0); k < ni; ++k )  temp[k] = p[k]; 
   }
   catch( bad_alloc& )
   {  delete[] temp;
      throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed ); }
   catch( TXptn& x )
   {
      delete[] temp; 
      if ( x.errorCode == TXptn::allocFailed )
         throw MyTmpltLibXptn( "renew", MyTmpltLibXptn::allocFailed);
      else
         throw;
   }
   delete[] p;   p = temp;
} // void renew
*/
template< typename T >
T** new2d( int nR, int nC )
{
   if ( nR <= 0 || nC <= 0 )
      throw MyTmpltLibXptn( "new2d",
                            MyTmpltLibXptn::paramError,
                            nR, nC );
   T** fv;
   try {  fv = new T*[nR];  }
   catch( bad_alloc& )
   {  throw MyTmpltLibXptn( "new2d",
                            MyTmpltLibXptn::allocFailed );  }
   for ( int r(0); r < nR; ++r )
   {
      try {  fv[r] = new T[nC];  }
      catch( bad_alloc& )
      {
         for ( int k(0); k < r; ++k )  delete[] fv[k];
         delete[] fv;
         throw MyTmpltLibXptn( "new2d",
                               MyTmpltLibXptn::allocFailed );
      } // catch
   } // for ( int r
   return fv;
} // new2d

template< typename T >
void delete2d( T**& a, int nR )
{
   if ( a == 0 && nR > 0 )
      throw MyTmpltLibXptn( "delete2d", MyTmpltLibXptn::noArray );
   for ( int r(0); r < nR; ++r )  delete[] a[r];
   delete[] a;  a = 0;
} // delete2d

// save1DTable --    binary   flNm  n 
//                  tbl.     
//                T::save( ostream& )
template < class T >
void save1DTable( string flNm, T tbl[], int n )
{
   if ( tbl == 0 && n > 0 ) 
      throw MyTmpltLibXptn( "save1DTable", MyTmpltLibXptn::noArray );
   ofstream bout( flNm.c_str(), ios_base::binary|ios_base::trunc );
   if ( bout.fail() )
      throw MyTmpltLibXptn( "save1DTable",
                            MyTmpltLibXptn::cannotCreate, flNm.c_str() );
   for ( int k(0); k < n; ++k )
   {
      tbl[k].save( bout );
      if ( bout.fail() )
         throw MyTmpltLibXptn( "save1DTable",
                               MyTmpltLibXptn::cannotWrite, flNm.c_str() );
   }
   bout.close();
} // save1DTable

#endif // _MYTMPLLIB_H
